@@ -6,17 +6,20 @@ This module implements a command line interface
 import argparse
 from cmd import Cmd
 from fnmatch import fnmatch
+import importlib.metadata
 import logging
 import os
 import shlex
 import stat
 import sys
-from typing import cast, Any, Dict, IO, List, NoReturn, Tuple, Union
+import traceback
+from typing import cast, Any, Callable, Dict, IO, List, NoReturn, Tuple, Union
 
 from duino_bus.dump_mem import dump_mem
 from duino_cli.colors import Color
 
 MAX_HISTORY_LINES: int = 40
+LOGGER = logging.getLogger(__name__)
 
 # argparse added exit_on_error support in 3.9
 if sys.version_info >= (3, 9, 0):
@@ -71,37 +74,6 @@ def add_arg(*args, **kwargs) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
     return (args, kwargs)
 
 
-def trim(docstring: str) -> str:
-    """Trims the leading spaces from docstring comments.
-
-    From http://www.python.org/dev/peps/pep-0257/
-
-    """
-    if not docstring:
-        return ''
-    # Convert tabs to spaces (following the normal Python rules)
-    # and split into a list of lines:
-    lines = docstring.expandtabs().splitlines()
-    # Determine minimum indentation (first line doesn't count):
-    indent = sys.maxsize
-    for line in lines[1:]:
-        stripped = line.lstrip()
-        if stripped:
-            indent = min(indent, len(line) - len(stripped))
-    # Remove indentation (first line is special):
-    trimmed = [lines[0].strip()]
-    if indent < sys.maxsize:
-        for line in lines[1:]:
-            trimmed.append(line[indent:].rstrip())
-    # Strip off trailing and leading blank lines:
-    while trimmed and not trimmed[-1]:
-        trimmed.pop()
-    while trimmed and not trimmed[0]:
-        trimmed.pop(0)
-    # Return a single string:
-    return '\n'.join(trimmed)
-
-
 class CommandLineOutput:
     """A class which allows easy integration of Cmd output into logging
     and also allows for easy capture of the output for testing purposes.
@@ -218,14 +190,10 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
     quitting = False
     output = None
 
-    def __init__(
-            self,
-            *args,
-            history_filename: str = '',
-            log=None,
-            filename=None,
-            **kwargs
-    ) -> None:
+    def __init__(self, params: Dict[str, Any], *args, log=None, filename=None, **kwargs) -> None:
+        self.params = params
+        self.bus = params['bus']
+        self.debug = params['debug']
         if 'stdin' in kwargs:
             Cmd.use_rawinput = False
         if not CommandLineBase.output:
@@ -247,8 +215,27 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
             readline.set_completer_delims(delims.replace("-", ""))
         self.redirect_filename = ''
         self.history = []
-        self.history_filename = history_filename
+        self.history_filename = params['history_filename']
         self.read_history()
+        self.plugins = {}
+        self.load_plugins()
+
+    def load_plugins(self) -> None:
+        """Loads plugins which have been installed."""
+        plugin_entry_points = importlib.metadata.entry_points()['duino_cli.plugin']
+        for plugin_entry_point in plugin_entry_points:
+            plugin_name = plugin_entry_point.name
+            LOGGER.info('Loading Plugin %s ...', plugin_name)
+            if self.debug:
+                LOGGER.info('  %s', plugin_entry_point)
+            try:
+                plugin_class = plugin_entry_point.load()
+                plugin = plugin_class(self)
+                self.plugins[plugin_name] = plugin
+            except Exception:  # pylint: disable=broad-exception-caught
+                LOGGER.error('Error encountered while loading plugin %s', plugin_name)
+                traceback.print_exc()
+        LOGGER.info('All plugins loaded')
 
     def add_completion_funcs(self, names, complete_func_name):
         """Helper function which adds a completion function for an array of
@@ -370,7 +357,17 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
             line = line[0:comment_idx]
             line = line.strip()
         try:
-            return Cmd.onecmd(self, line)
+            # We would normally use Cmd.onecmd here, but that doesn't work with
+            # plugins, so we duplicate the functionality here instead
+            cmd, arg, line = self.parseline(line)
+            if not line:
+                return self.emptyline()
+            if cmd is None or cmd == '':
+                return self.default(line)
+            fn = self.get_command(cmd)
+            if fn:
+                return fn(arg)
+            return self.default(line)
         except ValueError as err:
             return self.handle_exception(err)
 
@@ -459,13 +456,12 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
             args = parser.parse_args(args[1:])
         return args
 
-    def create_argparser(self, command):
+    def create_argparser(self, command) -> Union[None, CommandArgumentParser]:
         """Sets up and parses the command line if an argparse_xxx object exists."""
-        try:
-            argparse_args = getattr(self, "argparse_" + command)
-        except AttributeError:
+        argparse_args = self.get_command_args(command)
+        if not argparse_args:
             return None
-        doc_lines = getattr(self, "do_" + command).__doc__.expandtabs().splitlines()
+        doc_lines = self.get_command_help(command).expandtabs().splitlines()
         if '' in doc_lines:
             blank_idx = doc_lines.index('')
             usage = doc_lines[:blank_idx]
@@ -484,72 +480,34 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
             parser.add_argument(*args, **kwargs)
         return parser
 
-    argparse_help = (
-            add_arg(
-                    '-v',
-                    '--verbose',
-                    dest='verbose',
-                    action='store_true',
-                    help='Display more help for each command',
-                    default=False
-            ),
-            add_arg(
-                    'command',
-                    metavar='COMMAND',
-                    nargs='*',
-                    type=str,
-                    help='Command to get help on'
-            ),
-    )
-
     def get_commands(self) -> List[str]:
         """Gets a list of all of the commands."""
-        return [x[3:] for x in self.get_names() if x.startswith('do_')]
+        commands = []
+        for plugin in self.plugins.values():
+            commands.extend(plugin.get_commands())
+        return commands
 
-    def do_help(self, arg: str) -> Union[bool, None]:
-        """help [-v] [CMD]...
+    def get_command(self, command: str) -> Union[Callable, None]:
+        """Retrieves the function object associated with a command."""
+        for plugin in self.plugins.values():
+            cmd = plugin.get_command(command)
+            if cmd:
+                return cmd
+        return None
 
-           List available commands with "help" or detailed help with "help cmd".
-        """
-        # arg isn't really a string but since Cmd provides a do_help
-        # function we have to match the prototype.
-        args = cast(argparse.Namespace, arg)
-        if len(args.command) <= 0 and not args.verbose:
-            return Cmd.do_help(self, '') is True
-        if len(args.command) == 0:
-            help_cmd = ''
-        else:
-            help_cmd = args.command[0]
-        help_cmd = help_cmd.replace("-", "_")
+    def get_command_help(self, command: str) -> str:
+        """Retrieves the documentation associated with a command."""
+        fn = self.get_command(command)
+        if fn:
+            return fn.__doc__ or ''
+        return ''
 
-        if not help_cmd:
-            help_cmd = '*'
-
-        cmds = self.get_commands()
-        cmds.sort()
-
-        blank_line = False
-        for cmd in cmds:
-            if fnmatch(cmd, help_cmd):
-                if blank_line:
-                    print('--------------------------------------------------------------')
-                blank_line = True
-                parser = self.create_argparser(cmd)
-                if parser:
-                    # Need to figure out how to strip out the `usage:`
-                    # Need to figure out how to get indentation to work
-                    parser.print_help()
-                    continue
-
-                try:
-                    doc = getattr(self, 'do_' + cmd).__doc__
-                    if doc:
-                        doc = doc.format(command=cmd)
-                        self.stdout.write(f"{trim(str(doc))}\n")
-                        continue
-                except AttributeError:
-                    pass
-                self.stdout.write(f'{str(self.nohelp % (cmd,))}\n')
+    def get_command_args(self, command: str) -> Union[None, Tuple]:
+        """Retrievers the argparse arguments for a command."""
+        for plugin in self.plugins.values():
+            args = plugin.get_command_args(command)
+            if args:
+                return args
         return None
 
     def print(self, *args, end='\n', file=None) -> None:
@@ -560,6 +518,15 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
         file.write(line)
         file.flush()
 
+    def error(self, *args, end='\n', file=None) -> None:
+        """Like print, but allows for redirection."""
+        if file is None:
+            file = self.stdout
+        line = ' '.join(str(arg) for arg in args) + end
+
+        file.write(line)
+        file.flush()
+
     def dump_mem(self, buf, prefix='', addr=0) -> None:
         """Like print, but allows for redirection."""
         dump_mem(buf, prefix, addr, log=self.print)
@@ -583,31 +550,6 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
                 cmds[i] = cmd.replace("_", "-")
         Cmd.print_topics(self, header, cmds, cmdlen, maxcol)
 
-    def do_args(self, args: List[str]) -> Union[bool, None]:
-        """args [arguments...]
-
-           Debug function for verifying argument parsing. This function just
-           prints out each argument that it receives.
-        """
-        for idx, arg in enumerate(args):
-            self.print(f"arg[{idx}] = '{arg}'")
-
-    def do_exit(self, _) -> bool:
-        """exit
-
-           Exits from the program.
-        """
-        CommandLineBase.quitting = True
-        return True
-
-    def do_quit(self, _) -> bool:
-        """quit
-
-           Exits from the program.
-        """
-        CommandLineBase.quitting = True
-        return True
-
     def add_line_to_history(self, line: str) -> None:
         """Adds a line into the history."""
         prev_line = ''
@@ -654,14 +596,3 @@ class CommandLineBase(Cmd):  # pylint: disable=too-many-instance-attributes,too-
         for line in self.history:
             if fnmatch(line, history_filter):
                 self.print(line)
-
-
-def main():
-    """Test main."""
-    cli = CommandLineBase()
-    cli.cmdloop('')
-    cli.save_history()
-
-
-if __name__ == '__main__':
-    main()
